@import "kwantdoctheme.css";

/* spacing around blockquoted fields in parameters/attributes/returns */
/* Essential. Otherwise there is way too much space around*/
td.field-body > blockquote {
    margin-top: 0.1em;
    margin-bottom: 0.5em;
}

.field-list th {
  /* color: rgb(0,50,150); */
  /* background-color: #EEE8AA; */
  white-space: nowrap; /* Essential. Otherwise the colons can break
                          into a new line */
}

table.field-list {
  border-collapse: separate; /* Essential. Otherwise Parameters and Returns
                                are sharing one solid colored field. That looks
                                weird. */
  border-spacing: 10px;
  border-style: hidden;
}

div.specialnote-title {
    font-size: 105%;
    font-weight: bold;
    font-color: #3B4D3C;
    background-color: #DCE4DC;
    padding: 1em;
    padding-top: 0.4em;
    padding-bottom: 0.4em;
    margin-top: 10px;
    margin-bottom: 10px;
    border-width: 1px;
    border-color: #546C55;
    border-style: solid;
}

div.specialnote-body {
    background-color: #DCE4DC;
    padding: 1em;
    padding-top: 0.1em;
    padding-bottom: 0.4em;
    margin-top: -10px;
    margin-bottom: 10px;
    border-width: 1px;
    border-top-width: 0px;
    border-color: #546C55;
    border-style: solid;
}
