MathJax.Hub.Config({
    TeX: {
        Macros: {
            bra: ['{\\left\\langle #1\\right|}',1],
            ket: ['{\\left| #1\\right\\rangle}',1],
            braket: ['\\left\\langle#1|#2\\right\\rangle', 2],
            ri: '\\text{i}',
            rd: '\\text{d}'
        }
    }
});
